/*
 *
 * sample.c
 *
 *
 * Copyright 2008-2009 
 * Uchibori Laboratory, Numazu National College of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include "./head/motor.h"


/*
 * Protocol
 * # Ping
 * Req: 0x00
 * Res: 0x80
 *
 * # 制御パラメータの設定
 * Req: 0x01, 目標値, Kp, Ki
 * Res: 0x81
 *  目標値: -128?127(パルス/100ms)
 *  Kp: 0?255
 *  Ki: 0?255
 *
 * # モータの状態取得
 * Req: 0x02
 * Res: 0x82, 速度, PWMデューティ(2Byte), 累積カウント(4Byte)
 *  速度: -128?127[パルス/100ms]
 *  PWM duty: 0?255 (duty比: 0?100%)(要検討)
 *  累積カウント: -2^31?2^31-1[パルス]
 *
 * # スタート
 * Req: 0x04
 * Res: 0x84
 *
 * # ストップ
 * Req: 0x08
 * Res: 0x88
 *
 * # 制御パラメータの取得
 * Req: 0x10
 * Res: 0x90, 目標値, Kp, Ki
 *
 * # Unknown
 * Req: 0x??
 * Res: 0x40 | Req
 */
/*
#define MG3_MOTOR_PKT_PING  	0x00
#define MG3_MOTOR_PKT_SETCTL	0x01
#define MG3_MOTOR_PKT_GETST 	0x02
#define MG3_MOTOR_PKT_START 	0x04
#define MG3_MOTOR_PKT_STOP  	0x08
//#define MG3_MOTOR_PKT_GETCTL	0x10
#define MG3_MOTOR_PKT_ACK   	0x80



#define FNAME "/dev/mirS1"

// PI制御用の各ゲイン
#define KP 30
#define KI 5


void beep (int n)
{
	int i;

	for ( i = 0; i < n; i++ ) {
		fprintf(stderr, "\a");
		usleep(500000);
	}
}
*/

// モータと通信をするための基本ルーチン
// リクエストパケットreqをモータへ送り，返答パケットresを送る.
// それぞれのパケットの長さは，req_len，res_lenで指定する．
// 成功すれば，0が返り，失敗すると負数が返る．
// 失敗したときには，errnoにエラーコードがセットされる．
// reqパケットの内容が不正な時にはEBADRQCがセットされる.
// その他の引数が不正な場合，EINVALがセットされる.
// その他のエラーコードについては，read(2)，write(2)を参照．
int motor_request (int fd, const void *req, void *res, size_t req_len, size_t res_len)
{
	int err;
	uint8_t response;

	if ( req == NULL ) {
		errno = EINVAL;
		return -1;
	}

	if ( res == NULL ) {
		if ( res_len <= sizeof(response) ) {
			res = &response;
			res_len = sizeof(response);
		}
		else {
			errno = EINVAL;
			return -1;
		}
	}
	assert(res != NULL);


	// リクエストを送信する.
	err = write(fd, req, req_len);
	if ( err < 0 ) {
		err = errno;
		perror("Send");
		fprintf(stderr, "errno : %d\n", err);
		errno = err;
		return -1;
	}

	// 返答を受信する．
	err = read(fd, res, res_len);
	if ( err < 0 ) {
		err = errno;
		perror("Recv");
		errno = err;
		return -1;
	}

	// 返答がACKであるか確認する.
	if ( *(uint8_t *)res != (*(uint8_t *)req | MG3_MOTOR_PKT_ACK) ) {
		errno = EBADRQC; // 不正なリクエストコード
		perror("ReqCode");
		errno = EBADRQC; // 不正なリクエストコード
		return -1;
	}

	return 0;
}



// モータに制御パラメータを送るためのルーチン．
// オープンされたデバイスファイルfdより
// 設定したい速度speed，PI制御の各ゲインkp，kiを設定する．
// 返り値については，motor_request()を参照．
int motor_send_control_packet (int fd, uint8_t speed, uint8_t kp, uint8_t ki)
{
	struct {
		uint8_t command;
		uint8_t spped;
		uint8_t kp;
		uint8_t ki;
	} request = {MG3_MOTOR_PKT_SETCTL, speed, kp, ki};
	int err;

	// パケット長が4であるか確認する.
	assert(sizeof(request) == 4);

	// モータに制御値を送る．
	err = motor_request(fd, &request, NULL, sizeof(request), 0);

	return err;
}



// 現在にモータの状態を得るためのルーチン.
// オープンされたデバイスファイルfdより
// 回転速度speed，pwmのパルス幅，総回転数roundの値を取得する.
// 返り値については，motor_request()を参照．
int motor_recv_status_packet (int fd, int8_t *speed, uint16_t *pwm, int32_t *round)
{
	struct {
		uint8_t response;
		int8_t speed;
		uint16_t pwm;
		int32_t total_round;
	} response;
	int err;
	uint8_t request;

	// パケット長が8であるか確認する.
	assert(sizeof(response) == 8);

	// モータのステータスを取得する.
	request = MG3_MOTOR_PKT_GETST;
	err = motor_request(fd, &request, &response, sizeof(request), sizeof(response));
	if ( err < 0 ) {
		return err;
	}
	*speed = response.speed;
	*pwm   = response.pwm;
	*round = response.total_round;

	return err;
}



// モータに回転を開始させるルーチン
// 返り(ry
int motor_send_start_packet (int fd)
{
	uint8_t request;

	// モータの回転を開始する.
	request = MG3_MOTOR_PKT_START;
	return motor_request(fd, &request, NULL, 1, 0);
}

int motor_send_clear_packet (int fd)
{
	uint8_t request;

	// REカウント値のクリア
	request = MG3_MOTOR_PKT_PING;
	return motor_request(fd, &request, NULL, 1, 0);
}



// モータの回転を停止させるルーチン
// 返(ry
int motor_send_stop_packet (int fd)
{
	uint8_t request;

	// モータの回転を開始する.
	request = MG3_MOTOR_PKT_STOP;
	return motor_request(fd, &request, NULL, 1, 0);
}


/*
int main (void)
{
	int i;
	int fd;
	int8_t speed;         // 現在の速度(deg./100ms）
	uint16_t pwm;         // 現在のPWMのパルス幅．デューティ比はpwn/1024 未動作?
	int32_t total_round;  // 総回転角度(deg.)
	uint8_t request;

	// モータとの通信用デバイスを開く．
	fd = open(FNAME, O_RDWR);
	if ( fd < 0 ) {
		perror("Open");
		return -1;
	}

	// ぷっぷっ
	beep(2);

	// モータの回転を開始する.
	request = MG3_MOTOR_PKT_START;
	if ( motor_send_start_packet(fd) ) {
		perror("Send start");
		return -1;
	}


	// 　0．5秒間隔で10秒間，モータのレギュレータ制御を行う．
	for ( i = 0; i < 20; i++ ) {
		// ステータスを取得する．
		if ( motor_recv_status_packet(fd, &speed, &pwm, &total_round) < 0 ) {
			perror("Get motor status");
		}
		printf("speed : %d, pwm : %d(%x, %x), total : %ld[deg.]\n",
			(int)speed, (int)pwm, pwm & 0xff, pwm >> 8, (long)total_round);


		// 本当はここで，制御パラメータを計算する．



		// 目標値とゲインを設定
		if ( motor_send_control_packet(fd, 18, KP, KI) < 0 ) {
			perror("Set control parameter");
			return -1;
		}

		// 一休み
		usleep(500000);
	}

	// もう一度
	for ( i = 0; i < 20; i++ ) {
		// ステータスを取得する．
		if ( motor_recv_status_packet(fd, &speed, &pwm, &total_round) < 0 ) {
			perror("Get motor status");
		}
		printf("speed : %d, pwm : %d(%x, %x), total : %ld[deg.]\n",
			(int)speed, (int)pwm, pwm & 0xff, pwm >> 8, (long)total_round);


		// 本当はここで，制御パラメータを計算する．



		// 目標値とゲインを設定
		if ( motor_send_control_packet(fd, -18, KP, KI) < 0 ) {
			perror("Set control parameter");
			return -1;
		}

		// 一休み
		usleep(500000);
	}



	// ぷっぷっぷっ
	beep(3);

	// モータの回転を停止する.
	if ( motor_send_stop_packet(fd) < 0 ) {
		perror("Send stop");
		return -1;
	}


	close(fd);	

	return 0;
}
*/
